/* Copyright (c) 2013  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "Platform.h"
#include <cassert>
#include <cstdarg>
#include <cstdio>
#include <stdexcept>
#include <vector>
#include <windows.h>
#include <Poco/Debugger.h>
#include <Poco/Exception.h>
#include <Poco/Format.h>


static Debugger::PrintCallback _echo = NULL;


void Debugger::print(const std::string& msg)
{
	if (_echo) try { _echo(msg.c_str()); } catch (...) {}

	try
	{
		if (Poco::Debugger::isAvailable())
		{
			Poco::Debugger::message(msg);
		}
		else
		{
			std::wstring msg2;
			Platform::Charset::fromUTF8(msg, msg2);
			OutputDebugStringW(msg2.c_str());
		}
	}
	catch (...)
	{
		Poco::Debugger::enter(msg, __FILE__, __LINE__);
	}
}


void Debugger::printf(const char* const fmt, ...)
{
	try
	{
		va_list args;
		va_start(args, fmt);

		std::vector<char> buf(1024);
		const int len = _vsnprintf(&buf[0], buf.size(), fmt, args);

		va_end(args);

		if (len < 0 || len > (int) buf.size())
		{
			throw std::length_error(Poco::format(
				"vsnprintf returned %d", len).c_str());
		}

		print(std::string(&buf[0], len));
	}
	catch (...)
	{
		Poco::Debugger::enter(fmt, __FILE__, __LINE__);
	}
}


void Debugger::printException(const std::exception& except, const std::string& scope)
{
	std::string message(scope.empty() ? "" : (scope + " had exception: "));

	try
	{
		const Poco::Exception& ex = dynamic_cast<const Poco::Exception&>(except);
		message.append(ex.displayText());
	}
	catch (...)
	{
		message.append(except.what());
	}

	print(message);
}


void Debugger::setPrintCallback(const PrintCallback proc)
{
	_echo = proc;
}
