/* Copyright (c) 2013  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "ALACEncoder.h"
#include <cassert>


/** number of bytes in an ALAC packet header */
const size_t ALACEncoder::ALAC_HEADER_SIZE = 3;


/**
 * Builds ALAC-encoded audio data packet from specified raw audio data.
 *
 * @param packetBuffer ALAC audio data packet buffer
 * @param packetCapacity ALAC audio data packet buffer capacity
 * @param dataBuffer audio data buffer to encode and write to packet buffer
 * @param dataLength audio data buffer length
 */
void ALACEncoder::buildPacket(byte_t* const packetBuffer, const size_t packetCapacity,
	const byte_t* const dataBuffer, const size_t dataLength)
{
	assert(dataBuffer != NULL && dataLength > 0 && dataLength % 4 == 0);
	assert(packetBuffer != NULL && packetCapacity >= ALAC_HEADER_SIZE + dataLength);

	size_t bitOffset = 0;
	size_t byteOffset = 0;

	// channel count; 0 is mono, 1 is stereo
	alacEncode(packetBuffer, 1, 3, bitOffset, byteOffset);
	// unknown
	alacEncode(packetBuffer, 0, 4, bitOffset, byteOffset);
	// unknown
	alacEncode(packetBuffer, 0, 12, bitOffset, byteOffset);
	// has_size flag
	alacEncode(packetBuffer, 0, 1, bitOffset, byteOffset);
	// uknown
	alacEncode(packetBuffer, 0, 2, bitOffset, byteOffset);
	// no_compression flag
	alacEncode(packetBuffer, 1, 1, bitOffset, byteOffset);

	// encode audio data into network byte order
	for (size_t b = 0; b < dataLength; b += 2)
	{
		alacEncode(packetBuffer, dataBuffer[b + 1], 8, bitOffset, byteOffset);
		alacEncode(packetBuffer, dataBuffer[b], 8, bitOffset, byteOffset);
	}
}


/**
 * Performs ALAC encoding on specified value for specified number of bits.  The
 * resulting encoded data is appended to specified buffer at specified offset.
 *
 * @param buffer ALAC audio data packet buffer
 * @param value audio data value to encode and append to buffer
 * @param length audio data length
 * @param bitOffset ALAC packet bit offset
 * @param byteOffset ALAC packet byte offset
 */
void ALACEncoder::alacEncode(byte_t* const buffer, const int value, size_t length,
	size_t& bitOffset, size_t& byteOffset)
{
	static const byte_t masks[8] = { 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF };

	size_t writeLength;
	byte_t bitsToWrite;

	if (bitOffset != 0 && bitOffset + length > 8)
	{
		writeLength = 8 - bitOffset;
		bitsToWrite = static_cast<byte_t>(
			(value >> (length - writeLength)) << (8 - bitOffset - writeLength));
		buffer[byteOffset] |= bitsToWrite;
		length -= writeLength;
		bitOffset = 0;
		++byteOffset;
	}

	while (length >= 8)
	{
		bitsToWrite = static_cast<byte_t>((value >> (length - 8)) & 0xFF);
		buffer[byteOffset] |= bitsToWrite;
		length -= 8;
		bitOffset = 0;
		++byteOffset;
	}

	if (length > 0)
	{
		bitsToWrite = static_cast<byte_t>(
			(value & masks[length]) << (8 - bitOffset - length));
		buffer[byteOffset] |= bitsToWrite;
		bitOffset += length;
		if (bitOffset == 8)
		{
			bitOffset = 0;
			++byteOffset;
		}
	}
}
