/* Copyright (c) 2013  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "NumberParser.h"
#include "Platform.h"
#include "Plugin.h"
#include "Random.h"
#include "RAOPDefs.h"
#include "RTSPClient.h"
#include <cassert>
#include <cstring>
#include <stdexcept>
#include <string>
#include <vector>
#include <openssl/evp.h>
#include <openssl/md5.h>
#include <openssl/rsa.h>
#include <Poco/Format.h>
#include <Poco/Timespan.h>


using Poco::Timespan;
using Poco::Net::StreamSocket;


const int RTSP_STATUS_CODE_OK           = 200;
const int RTSP_STATUS_CODE_UNAUTHORIZED = 401;


static const std::string AUDIO_JACK_STATUS_HEADER("Audio-Jack-Status");
static const std::string AUDIO_LATENCY_HEADER("Audio-Latency");
static const std::string CHALLENGE_RESPONSE_HEADER("Apple-Response");
static const std::string SESSION_HEADER("Session");
static const std::string TRANSPORT_HEADER("Transport");
static const std::string WWW_AUTHENTICATE_HEADER("WWW-Authenticate");


//------------------------------------------------------------------------------


RTSPClient::RTSPClient(StreamSocket& socket, const uint32_t& remoteControlId)
:
	_teardownRequired(false),
	_messageSequenceNumber(0),
	_localSessionId(),
	_remoteSessionId(),
	_remoteControlId(remoteControlId),
	_authenticationCasing(0),
	_rtspSocket(socket)
{
	_rtspSocket.setBlocking(true);
	_rtspSocket.setKeepAlive(true);
	_rtspSocket.setLinger(true, 1);
	_rtspSocket.setNoDelay(true);

	_rtspSocket.setSendTimeout(Timespan(10, 0));
	_rtspSocket.setReceiveTimeout(Timespan(10, 0));
}


RTSPClient::~RTSPClient()
{
	try
	{
		_rtspSocket.close();
	}
	CATCH_ALL
}


/**
 * Sends RTSP OPTIONS message.
 *
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doOptions(void* const rsaKey)
{
	buffer_t challengeNonce(16);
	std::string requestHead;

	if (rsaKey)
	{
		// generate random challenge nonce
		Random::fill(&challengeNonce[0], challengeNonce.size());

		// base64 encode challenge nonce
		std::vector<char> encodedChallengeNonce(32);
		const int encodedChallengeNonceLength = EVP_EncodeBlock(
			(unsigned char*) &encodedChallengeNonce[0],
			(unsigned char*) &challengeNonce[0],
			(int) challengeNonce.size());
		if (encodedChallengeNonceLength <= 0
			|| encodedChallengeNonceLength > (int) encodedChallengeNonce.size())
		{
			throw std::runtime_error("EVP_EncodeBlock failed");
		}
		encodedChallengeNonce.resize(encodedChallengeNonceLength);

		// remove padding from base64-encoded challenge nonce
		while (encodedChallengeNonce.size() > 0
			&& encodedChallengeNonce.back() == '=')
		{
			encodedChallengeNonce.pop_back();
		}

		requestHead.append(Poco::format("Apple-Challenge: %s\r\n",
			std::string(&encodedChallengeNonce[0], encodedChallengeNonce.size())));
	}

	Response response;
	sendRequestReceiveResponse("OPTIONS", requestHead, "", response);

	if (response.statusCode != RTSP_STATUS_CODE_OK
		// allow unauthorized condition to flow through so challenge is checked
		&& response.statusCode != RTSP_STATUS_CODE_UNAUTHORIZED)
	{
		return response.statusCode;
	}

	if (rsaKey)
	{
		if (!response.headers.count(CHALLENGE_RESPONSE_HEADER))
		{
			return -200000;
		}
		std::string encodedResponseNonce(response.headers[CHALLENGE_RESPONSE_HEADER]);

		// add padding to base64-encoded string as necessary
		while (encodedResponseNonce.length() % 4 != 0)
		{
			encodedResponseNonce.push_back('=');
		}

		// base64 decode RSA-encrypted response nonce
		buffer_t encryptedResponseNonce(
			buffer_t::size_type(encodedResponseNonce.length() * 0.75));
		const int encryptedResponseNonceLength = EVP_DecodeBlock(
			(unsigned char*) &encryptedResponseNonce[0],
			(unsigned char*) encodedResponseNonce.c_str(),
			(int) encodedResponseNonce.length());
		if (encryptedResponseNonceLength <= 0
			|| encryptedResponseNonceLength > (int) encryptedResponseNonce.size())
		{
			throw std::runtime_error("EVP_DecodeBlock failed");
		}

		assert(RSA_size((RSA*) rsaKey) == 256);

		// remove padding from RSA-encrypted response nonce
		while (encryptedResponseNonce.size() > 256
			&& encryptedResponseNonce.back() == 0)
		{
			encryptedResponseNonce.pop_back();
		}

		// recheck length of RSA-encrypted response nonce
		if (encryptedResponseNonce.size() != 256)
		{
			return -200001;
		}

		// RSA decrypt response nonce
		buffer_t responseNonce(256);
		const int responseNonceLength = RSA_public_decrypt(
			(int) encryptedResponseNonce.size(),
			(unsigned char*) &encryptedResponseNonce[0],
			(unsigned char*) &responseNonce[0],
			(RSA*) rsaKey,
			RSA_PKCS1_PADDING);
		if (responseNonceLength <= 0
			|| responseNonceLength > (int) responseNonce.size())
		{
			return -200002;
		}

		if (responseNonceLength < (int) challengeNonce.size())
		{
			return -200003;
		}

		// compare response nonce to challenge nonce; they should match if
		// remote speakers posseses private counterpart to RSA public key used
		if (std::memcmp(
			&responseNonce[0], &challengeNonce[0], challengeNonce.size()) != 0)
		{
			return -200004;
		}
	}

	return response.statusCode;
}


/**
 * Sends RTSP ANNOUNCE message.
 *
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doAnnounce(const std::string& aesKey, const std::string& aesIV)
{
	// generate local session identifier
	Random::fill(&_localSessionId, sizeof(uint32_t));

	std::string requestBody(Poco::format(
		"v=0\r\n"
		"o=iTunes %u 0 IN IP4 %s\r\n"
		"s=iTunes\r\n"
		"c=IN IP4 %s\r\n"
		"t=0 0\r\n",
		_localSessionId,
		_rtspSocket.address().host().toString(),
		_rtspSocket.peerAddress().host().toString()));

	requestBody.append(Poco::format(
		"m=audio 0 RTP/AVP 96\r\n"
		"a=rtpmap:96 AppleLossless\r\n"
		"a=fmtp:96 %u 0 %u 40 10 14 %u 255 0 0 %u\r\n",
		RAOP_PACKET_MAX_SAMPLES_PER_CHANNEL,
		RAOP_BITS_PER_SAMPLE,
		RAOP_CHANNEL_COUNT,
		RAOP_SAMPLES_PER_SECOND));

	if (!aesKey.empty() && !aesIV.empty())
	{
		requestBody.append(Poco::format(
			"a=rsaaeskey:%s\r\n"
			"a=aesiv:%s\r\n",
			aesKey, aesIV));
	}

	const std::string requestHead(Poco::format(
		"Content-Type: application/sdp\r\n"
		"Content-Length: %u\r\n",
		requestBody.length()));

	Response response;
	sendRequestReceiveResponse("ANNOUNCE", requestHead, requestBody, response);

	return response.statusCode;
}


/**
 * Sends RTSP SETUP message.
 *
 * @param audioLatency [out] remote speakers audio latency (in number of samples)
 * @param audioJackStatus [out] status of remote speakers audio jack
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doSetup(uint16_t& serverPort, uint16_t& controlPort, uint16_t& timingPort,
	unsigned int& audioLatency, AudioJackStatus& audioJackStatus)
{
	const std::string requestHead(Poco::format(
		"Transport: RTP/AVP/UDP;unicast;interleaved=0-1;mode=record;"
		"control_port=%hu;timing_port=%hu\r\n", controlPort, timingPort));

	Response response;
	sendRequestReceiveResponse("SETUP", requestHead, "", response);

	if (response.statusCode != RTSP_STATUS_CODE_OK)
	{
		return response.statusCode;
	}

	// check for session header
	if (!response.headers.count(SESSION_HEADER))
	{
		return -200100;
	}
	const std::string& sessionHeader(response.headers[SESSION_HEADER]);

	// validate session header
	if (sessionHeader.empty())
	{
		return -200101;
	}
	_remoteSessionId = sessionHeader;

	// check for transport header
	if (!response.headers.count(TRANSPORT_HEADER))
	{
		return -200102;
	}
	const std::string& transportHeader(response.headers[TRANSPORT_HEADER]);

	// parse transport header
	std::string::size_type beg = 0;
	while (beg != std::string::npos)
	{
		std::string::size_type mid = transportHeader.find_first_of('=', beg);
		std::string::size_type end = transportHeader.find_first_of(';', beg);

		if (mid != std::string::npos && mid < end)
		{
			const std::string key(transportHeader.substr(beg, mid - beg));
			++mid; // advance past '='
			const std::string val(transportHeader.substr(mid, end - mid));

			if (key == "server_port")
			{
				serverPort = NumberParser::parseDecimalIntegerTo<uint16_t>(val);
			}
			else if (key == "control_port")
			{
				controlPort = NumberParser::parseDecimalIntegerTo<uint16_t>(val);
			}
			else if (key == "timing_port")
			{
				timingPort = NumberParser::parseDecimalIntegerTo<uint16_t>(val);
			}
		}

		beg = transportHeader.find_first_not_of(';', end);
	}

	// check for audio latency header
	if (response.headers.count(AUDIO_LATENCY_HEADER))
	{
		const std::string& audioLatencyHeader(
			response.headers[AUDIO_LATENCY_HEADER]);

		// parse audio latency header
		audioLatency =
			NumberParser::parseDecimalIntegerTo<unsigned int>(audioLatencyHeader);
	}

	// check for audio jack status header
	if (response.headers.count(AUDIO_JACK_STATUS_HEADER))
	{
		const std::string& audioJackStatusHeader(
			response.headers[AUDIO_JACK_STATUS_HEADER]);

		// parse audio jack status header
		audioJackStatus = (audioJackStatusHeader == "disconnected"
			? AUDIO_JACK_DISCONNECTED : AUDIO_JACK_CONNECTED);
	}

	return response.statusCode;
}


/**
 * Sends RTSP RECORD message.
 *
 * @param audioLatency [out] remote speakers audio latency (in number of samples)
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doRecord(const uint16_t rtpSeqNum, const uint32_t rtpTime,
	unsigned int& audioLatency)
{
	const std::string requestHead(Poco::format(
		"Session: %s\r\n"
		"Range: npt=0-\r\n"
		"RTP-Info: seq=%hu;rtptime=%u\r\n",
		_remoteSessionId, rtpSeqNum, rtpTime));

	Response response;
	sendRequestReceiveResponse("RECORD", requestHead, "", response);

	if (response.statusCode != RTSP_STATUS_CODE_OK)
	{
		return response.statusCode;
	}

	// check for audio latency header
	if (response.headers.count(AUDIO_LATENCY_HEADER))
	{
		const std::string& audioLatencyHeader(
			response.headers[AUDIO_LATENCY_HEADER]);

		// parse audio latency header
		audioLatency =
			NumberParser::parseDecimalIntegerTo<unsigned int>(audioLatencyHeader);
	}

	// indicate that teardown is required on disconnect because record message
	// was received successfully and a session is now fully established
	_teardownRequired = true;

	return response.statusCode;
}


/**
 * Sends RTSP FLUSH message.
 *
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doFlush(const uint16_t rtpSeqNum, const uint32_t rtpTime)
{
	const std::string requestHead(Poco::format(
		"Session: %s\r\n"
		"RTP-Info: seq=%hu;rtptime=%u\r\n",
		_remoteSessionId, rtpSeqNum, rtpTime));

	Response response;
	sendRequestReceiveResponse("FLUSH", requestHead, "", response);

	return response.statusCode;
}


/**
 * Sends RTSP TEARDOWN message.
 *
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doTeardown()
{
	if (!_teardownRequired)
	{
		return RTSP_STATUS_CODE_OK;
	}

	const std::string requestHead(Poco::format(
		"Session: %s\r\n", _remoteSessionId));

	Response response;
	sendRequestReceiveResponse("TEARDOWN", requestHead, "", response);

	return response.statusCode;
}


/**
 * Sends RTSP SET_PARAMETER message.
 *
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doSetParameter(
	const std::string& parameterName, const std::string& parameterValue)
{
	assert(!parameterName.empty());
	assert(!parameterValue.empty());

	const std::string requestBody(Poco::format(
		"%s: %s\r\n", parameterName, parameterValue));

	const std::string requestHead(Poco::format(
		"Session: %s\r\n"
		"Content-Type: text/parameters\r\n"
		"Content-Length: %u\r\n",
		_remoteSessionId, requestBody.length()));

	Response response;
	sendRequestReceiveResponse("SET_PARAMETER", requestHead, requestBody, response);

	return response.statusCode;
}


/**
 * Sends RTSP SET_PARAMETER message.
 *
 * @return zero on success; non-zero on failure
 */
int RTSPClient::doSetParameter(const std::string& contentType,
	const buffer_t& requestBody, const uint32_t rtpTime)
{
	const uint32_t contentLength = requestBody.size();

	const std::string requestHead(Poco::format(
		"Session: %s\r\n"
		"Content-Type: %s\r\n"
		"Content-Length: %u\r\n"
		"RTP-Info: rtptime=%u\r\n",
		_remoteSessionId, contentType, contentLength, rtpTime));

	std::string requestData(contentLength == 0 ? ""
		: reinterpret_cast<const char*>(&requestBody[0]), requestBody.size());

	Response response;
	sendRequestReceiveResponse("SET_PARAMETER", requestHead, requestData, response);

	return response.statusCode;
}


/**
 * Sends specified RTSP message to remote speakers associated with specified
 * session and retrieves response.  This function requires specified request
 * method to construct the base request headers.  It will also use specified
 * additional headers and specified optional body.  If a body is to be included
 * with the request it is assumed that the caller has included appropriate
 * content type and content length headers.
 *
 * @param requestMethod RTSP request method
 * @param requestHead additional RTSP request headers
 * @param requestBody optional RTSP request body
 * @param response RTSP response structure
 */
void RTSPClient::sendRequestReceiveResponse(const std::string& requestMethod,
	const std::string& requestHead, const std::string& requestBody, Response& response)
{
	const std::string requestText(buildRequest(requestMethod, requestHead, requestBody));
	Debugger::print(requestText + std::string(80, '-'));

	sendRequest(requestText);

	const std::string responseText(receiveResponse());
	Debugger::print(responseText + std::string(80, '-'));

	// increment sequence number after successful reception of response
	_messageSequenceNumber += 1;

	parseResponse(responseText, response);

	// validate response sequence number
	if (response.headers.count("CSeq"))
	{
		const std::string& cSeqHeader(response.headers["CSeq"]);
		assert(_messageSequenceNumber
			== NumberParser::parseDecimalIntegerTo<uint32_t>(cSeqHeader));
	}

	// check response for authenticate header
	if (response.headers.count(WWW_AUTHENTICATE_HEADER))
	{
		parseAuthenticateHeader(response.headers[WWW_AUTHENTICATE_HEADER]);
	}
}


std::string RTSPClient::buildRequest(const std::string& requestMethod,
	const std::string& requestHead, const std::string& requestBody) const
{
	assert(!requestMethod.empty());

	// build request URI
	const std::string requestURI(_localSessionId == 0 ? "*" : Poco::format(
		"rtsp://%s/%u", _rtspSocket.address().host().toString(), _localSessionId));

	std::string requestText;
	requestText.append(Poco::format("%s %s RTSP/1.0\r\n", requestMethod, requestURI));
	requestText.append(Poco::format("CSeq: %u\r\n", _messageSequenceNumber + 1));
	requestText.append(Poco::format("User-Agent: %s\r\n", Plugin::userAgent()));
	requestText.append(Poco::format("Client-Instance: %016?X\r\n", Plugin::dacpId()));
	requestText.append(Poco::format("DACP-ID: %016?X\r\n", Plugin::dacpId()));
	requestText.append(Poco::format("Active-Remote: %u\r\n", _remoteControlId));
	if (!_authenticationMethod.empty())
		requestText.append(buildAuthorizationHeader(requestMethod, requestURI));
	requestText.append(requestHead);
	requestText.append("\r\n");
	requestText.append(requestBody);

	return requestText;
}


void RTSPClient::sendRequest(const std::string& requestText)
{
	assert(!requestText.empty());

	const char* const requestBuffer = requestText.c_str();
	const size_t requestLength = requestText.length();

	size_t bytesSent = 0;
	while (bytesSent < requestLength)
	{
		const int returnCode = _rtspSocket.sendBytes(requestBuffer + bytesSent,
			requestLength - bytesSent);
		if (returnCode <= 0)
		{
			throw std::runtime_error(Poco::format(
				"_rtspSocket.sendBytes returned %i", returnCode));
		}

		bytesSent += static_cast<size_t>(returnCode);
	}
	if (bytesSent != requestLength)
	{
		throw std::runtime_error("bytesSent != requestLength");
	}
}


std::string RTSPClient::receiveResponse()
{
	std::vector<char> responseBuffer(8192);
	size_t responseLength = 0;

	do
	{
		const int returnCode = _rtspSocket.receiveBytes(
			&responseBuffer[responseLength],
			responseBuffer.size() - responseLength);
		if (returnCode <= 0)
		{
			throw std::runtime_error(Poco::format(
				"_rtspSocket.receiveBytes returned %i", returnCode));
		}

		responseLength += static_cast<size_t>(returnCode);
	}
	while (std::strncmp(&responseBuffer[responseLength - 4], "\r\n\r\n", 4) != 0);

	assert(responseLength < responseBuffer.size());
	return std::string(&responseBuffer[0], responseLength);
}


void RTSPClient::parseResponse(const std::string& responseText, Response& response)
{
	assert(!responseText.empty());

	std::string::size_type beg = 0;
	std::string::size_type end;

	// parse response for protocol
	end = responseText.find_first_of("/", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.protocol = responseText.substr(beg, end - beg);
	beg = responseText.find_first_not_of("/", end);

	// parse response for version
	end = responseText.find_first_of(" ", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.version = responseText.substr(beg, end - beg);
	beg = responseText.find_first_not_of(" ", end);

	// parse response for status code
	end = responseText.find_first_of(" ", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.statusCode = NumberParser::parseDecimalIntegerTo<int>(
		responseText.substr(beg, end - beg));
	beg = responseText.find_first_not_of(" ", end);

	// parse response for status text
	end = responseText.find_first_of("\r\n", beg);
	if (end <= beg || end == std::string::npos)
	{
		throw std::invalid_argument("responseText");
	}
	response.statusText = responseText.substr(beg, end - beg);
	beg = responseText.find_first_not_of("\r\n", end);

	// parse response headers
	while (beg != std::string::npos)
	{
		assert(end != std::string::npos);

		// tokenize header into name and value
		end = responseText.find_first_of(":", beg);
		const std::string name = responseText.substr(beg, end - beg);
		beg = responseText.find_first_not_of(": ", end);

		end = responseText.find_first_of("\r\n", beg);
		const std::string value = responseText.substr(beg, end - beg);
		beg = responseText.find_first_not_of("\r\n", end);

		response.headers.insert(std::make_pair(name, value));
	}
}


void RTSPClient::parseAuthenticateHeader(const std::string& authenticateHeader)
{
	assert(!authenticateHeader.empty());

	std::string::size_type beg = 0;
	std::string::size_type end = authenticateHeader.find_first_of(" ", beg);

	if (end <= beg || end == std::string::npos)
		throw std::runtime_error("end <= beg || end == std::string::npos");

	_authenticationMethod = authenticateHeader.substr(beg, end - beg);

	if (_authenticationMethod != "Digest")
		throw std::runtime_error("_authenticationMethod != \"Digest\"");

	beg = authenticateHeader.find_first_not_of(" ", end);

	while (beg != std::string::npos)
	{
		assert(end != std::string::npos);

		end = authenticateHeader.find_first_of("=\"", beg);
		const std::string name(authenticateHeader.substr(beg, end - beg));
		beg = authenticateHeader.find_first_not_of("=\"", end);

		end = authenticateHeader.find_first_of("\"", beg);
		const std::string value(authenticateHeader.substr(beg, end - beg));
		beg = authenticateHeader.find_first_not_of("\", ", end);

		if (name == "realm")
		{
			assert(!value.empty());
			_authenticationRealm = value;
		}
		else if (name == "nonce")
		{
			assert(!value.empty());
			_authenticationNonce = value;
		}
	}

	if (_authenticationRealm.empty() || _authenticationNonce.empty())
		throw std::runtime_error(
			"_authenticationRealm.empty() || _authenticationNonce.empty()");

	_authenticationCasing =
		(_authenticationNonce.find_first_of("ABCDEF") != std::string::npos ? 1 : -1);
}


#define HEX(bytes, length, string, capacity, casing)                           \
{                                                                              \
	const char *format = (casing < 0 ? "%.2x" : casing > 0 ? "%.2X" : NULL);   \
	for (size_t i = 0; i < length; ++i)                                        \
		sprintf_s(string + 2 * i, capacity - 2 * i, format, bytes[i]);         \
}


std::string RTSPClient::buildAuthorizationHeader(
	const std::string& requestMethod, const std::string& requestURI) const
{
	int length;
	char temp[256];
	byte_t ha1Digest[MD5_DIGEST_LENGTH];
	char ha1String[MD5_DIGEST_LENGTH * 2 + 1];
	byte_t ha2Digest[MD5_DIGEST_LENGTH];
	char ha2String[MD5_DIGEST_LENGTH * 2 + 1];
	byte_t responseDigest[MD5_DIGEST_LENGTH];
	char responseString[MD5_DIGEST_LENGTH * 2 + 1];

	assert(!requestMethod.empty());
	assert(!requestURI.empty());

	// this function assumes RFC 2617 digest authentication
	assert(_authenticationMethod == "Digest");

	// concatenate username, realm and password
	length = sprintf_s(temp, sizeof(temp), "iTunes:%s:%s",
		_authenticationRealm.c_str(), _authenticationPassword.c_str());
	assert(length > 0 && length < sizeof(temp));

	// create HA1 digest
	MD5((unsigned char*) temp, length, ha1Digest);
	HEX(ha1Digest, sizeof(ha1Digest), ha1String, sizeof(ha1String),
		_authenticationCasing);

	// concatenate request method and URI
	length = sprintf_s(temp, sizeof(temp), "%s:%s",
		requestMethod.c_str(), requestURI.c_str());
	assert(length > 0 && length < sizeof(temp));

	// create HA2 digest
	MD5((unsigned char*) temp, length, ha2Digest);
	HEX(ha2Digest, sizeof(ha2Digest), ha2String, sizeof(ha2String),
		_authenticationCasing);

	// concatenate HA1 digest, nonce and HA2 digest
	length = sprintf_s(temp, sizeof(temp), "%s:%s:%s",
		ha1String, _authenticationNonce.c_str(), ha2String);
	assert(length > 0 && length < sizeof(temp));

	// create response digest
	MD5((unsigned char*) temp, length, responseDigest);
	HEX(responseDigest, sizeof(responseDigest), responseString,
		sizeof(responseString), _authenticationCasing);

	// build request authorization header
	length = sprintf_s(temp, sizeof(temp),
		"Authorization: %s username=\"iTunes\", realm=\"%s\", nonce=\"%s\", "
		"uri=\"%s\", response=\"%s\"\r\n",
		_authenticationMethod.c_str(), _authenticationRealm.c_str(),
		_authenticationNonce.c_str(), requestURI.c_str(), responseString);
	assert(length > 0 && length < sizeof(temp));

	return temp;
}
