/* Copyright (c) 2013  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "DeviceDiscovery.h"
#include "DeviceInfo.h"
#include "Options.h"
#include "OptionsSubmenu.h"
#include "Platform.h"
#include "Plugin.h"
#include "Resources.h"
#include <cassert>
#include <cstring>
#include <string>
#include <vector>


#define LABEL_SUBMENU "Remote Speakers"
#define LABEL_DUMYITM "(none found)"

static const UINT ID_SUBMENU = Plugin::id();
static const BYTE ID_SUBMENU_ITEM = 222;


class OptionsSubmenuImpl
:
	public DeviceDiscovery::Listener,
	private Uncopyable
{
	friend class OptionsSubmenu;

	static LRESULT CALLBACK getMessageProc(int, WPARAM, LPARAM);
	static LRESULT CALLBACK callWindowProc(int, WPARAM, LPARAM);
	static OptionsSubmenuImpl* _this;

	explicit OptionsSubmenuImpl(HWND);
	~OptionsSubmenuImpl();

	LRESULT processMessage(UINT, WPARAM, LPARAM);
	void onInitMenuPopup(HMENU, int);
	void onUninitMenuPopup(HMENU);
	void onItemSelection(WPARAM);

	void onDeviceFound(const DeviceInfo&);
	void onDeviceLost(const DeviceInfo&);

	void insertSubmenuItem(const DeviceInfo&, bool, bool = false);
	bool removeSubmenuItem(const DeviceInfo&);
	void populateSubmenu(const DeviceInfoSet&);
	void enumerateSubmenuItems(Options::SharedPtr) const;
	void finalizeSubmenu();

	Options::SharedPtr buildOptions();

	HWND _playerWindow;
	HHOOK _getMessageProcHook;
	HHOOK _callWindowProcHook;
	HMENU _submenu;
	BYTE _itemCount;
};


//------------------------------------------------------------------------------


OptionsSubmenu::OptionsSubmenu(HWND playerWindow)
:
	_impl(new OptionsSubmenuImpl(playerWindow))
{
}


OptionsSubmenu::~OptionsSubmenu()
{
	delete _impl;
}


//------------------------------------------------------------------------------


OptionsSubmenuImpl* OptionsSubmenuImpl::_this = NULL;


LRESULT CALLBACK OptionsSubmenuImpl::getMessageProc(const int code,
	WPARAM wParam, LPARAM lParam)
{
	LPMSG msg;

	if (code < 0)
	{
		return CallNextHookEx(_this->_getMessageProcHook, code, wParam, lParam);
	}

	msg = (LPMSG) lParam;

	// verify event is associated with player playerWindow before processing
	if (msg->hwnd == _this->_playerWindow
		&& _this->processMessage(msg->message, msg->wParam, msg->lParam) == 0)
	{
		return 0;
	}

	return CallNextHookEx(_this->_getMessageProcHook, code, wParam, lParam);
}


LRESULT CALLBACK OptionsSubmenuImpl::callWindowProc(const int code,
	WPARAM wParam, LPARAM lParam)
{
	LPCWPSTRUCT msg;

	if (code < 0)
	{
		return CallNextHookEx(_this->_callWindowProcHook, code, wParam, lParam);
	}

	msg = (LPCWPSTRUCT) lParam;

	// verify association with player playerWindow before processing
	if (msg->hwnd == _this->_playerWindow
		&& _this->processMessage(msg->message, msg->wParam, msg->lParam) == 0)
	{
		return 0;
	}

	return CallNextHookEx(_this->_callWindowProcHook, code, wParam, lParam);
}


OptionsSubmenuImpl::OptionsSubmenuImpl(HWND playerWindow)
:
	_playerWindow(playerWindow),
	_getMessageProcHook(0),
	_callWindowProcHook(0),
	_submenu(NULL),
	_itemCount(0)
{
	OptionsSubmenuImpl::_this = this;

	// hook into get message chain
	_getMessageProcHook = SetWindowsHookEx(WH_GETMESSAGE,
		OptionsSubmenuImpl::getMessageProc, NULL, GetCurrentThreadId());
	if (_getMessageProcHook == NULL)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	// hook into call window chain
	_callWindowProcHook = SetWindowsHookEx(WH_CALLWNDPROC,
		OptionsSubmenuImpl::callWindowProc, NULL, GetCurrentThreadId());
	if (_callWindowProcHook == NULL)
	{
		DWORD lastError = GetLastError();
		UnhookWindowsHookEx(_getMessageProcHook);
		assert(lastError == 0);
	}
}


OptionsSubmenuImpl::~OptionsSubmenuImpl()
{
	// unhook from get message chain
	if (UnhookWindowsHookEx(_getMessageProcHook) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	// unhook from call playerWindow chain
	if (UnhookWindowsHookEx(_callWindowProcHook) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	// clean up dynamically allocated resources
	finalizeSubmenu();
}


LRESULT OptionsSubmenuImpl::processMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_INITMENUPOPUP:
	case WM_UNINITMENUPOPUP:
		{
			HMENU menu = (HMENU) wParam;
			HMENU mainMenu = (HMENU) SendMessage(_playerWindow, WM_USER, 0, 281);
			HMENU optionsMenu = (HMENU) SendMessage(_playerWindow, WM_USER, 3, 281);

			// verify association with correct player menu before processing
			if (menu == mainMenu || menu == optionsMenu)
			{
				if (message == WM_INITMENUPOPUP)
				{
					onInitMenuPopup(menu,
						GetMenuItemCount(menu) - (menu == mainMenu ? 5 : 2));
				}
				else if (message == WM_UNINITMENUPOPUP)
				{
					onUninitMenuPopup(menu);
				}
			}
		}
		break;

	case WM_COMMAND:
	case WM_SYSCOMMAND:
		// verify association with this submenu before processing
		if (HIWORD(wParam) == 0 && LOBYTE(LOWORD(wParam)) == ID_SUBMENU_ITEM)
		{
			onItemSelection(wParam);
		}
		break;
	}

	return TRUE;
}


void OptionsSubmenuImpl::onInitMenuPopup(HMENU parentMenu, const int index)
{
	// destroy old submenu
	finalizeSubmenu();

	// create new submenu
	_submenu = CreatePopupMenu();
	if (_submenu == NULL)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	// install submenu item in player's menu
	MENUITEMINFO item;
	std::memset(&item, 0, sizeof(MENUITEMINFO));
	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_ID | MIIM_FTYPE | MIIM_STRING | MIIM_STATE | MIIM_SUBMENU;
	item.wID = ID_SUBMENU;
	item.fType = MFT_STRING;
	item.dwTypeData = TEXT(LABEL_SUBMENU);
	item.fState = MFS_ENABLED;
	item.hSubMenu = _submenu;
	if (InsertMenuItem(parentMenu, index, TRUE, &item) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	// append dummy item to submenu
	std::memset(&item, 0, sizeof(MENUITEMINFO));
	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_STATE;
	item.fType = MFT_STRING;
	item.dwTypeData = TEXT(LABEL_DUMYITM);
	item.fState = MFS_DISABLED;
	if (InsertMenuItem(_submenu, GetMenuItemCount(_submenu), TRUE, &item) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	// hold reference to active options
	const Options::SharedPtr options = Options::getOptions();

	populateSubmenu(options->devices());

	DeviceDiscovery::browseDevices(*this);
}


void OptionsSubmenuImpl::onUninitMenuPopup(HMENU parentMenu)
{
	DeviceDiscovery::stopBrowsing(*this);

	if (RemoveMenu(parentMenu, ID_SUBMENU, MF_BYCOMMAND) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}
}


void OptionsSubmenuImpl::onItemSelection(WPARAM wParam)
{
	// toggle checked state of item
	MENUITEMINFO item;
	std::memset(&item, 0, sizeof(MENUITEMINFO));
	item.cbSize = sizeof(MENUITEMINFO);
	item.fMask = MIIM_STATE | MIIM_DATA;
	if (GetMenuItemInfo(_submenu, (UINT) wParam, FALSE, &item) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}
	item.fState ^= MFS_CHECKED;
	if (SetMenuItemInfo(_submenu, (UINT) wParam, FALSE, &item) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	Options::setOptions(buildOptions());
}


void OptionsSubmenuImpl::onDeviceFound(const DeviceInfo& device)
{
	const bool checked = removeSubmenuItem(device);
	insertSubmenuItem(device, checked);
}


void OptionsSubmenuImpl::onDeviceLost(const DeviceInfo& device)
{
	const bool checked = removeSubmenuItem(device);
	if (checked)
	{
		// insert a stale item to retain visibility of checked item
		insertSubmenuItem(device, true, true);
	}
}


//------------------------------------------------------------------------------


void OptionsSubmenuImpl::insertSubmenuItem(const DeviceInfo& device,
	const bool check, const bool stale)
{
	int index, count;
	MENUITEMINFO item;

	// search submenu for dummy item
	for (index = 0, count = GetMenuItemCount(_submenu); index < count; ++index)
	{
		std::memset(&item, 0, sizeof(MENUITEMINFO));
		item.cbSize = sizeof(MENUITEMINFO);
		item.fMask = MIIM_STRING;

		if (GetMenuItemInfo(_submenu, (UINT) index, TRUE, &item) == 0)
		{
			DWORD lastError = GetLastError();
			assert(lastError == 0);
		}

		std::vector<char_t> stringBuffer(++item.cch);
		item.dwTypeData = (LPTSTR) &stringBuffer[0];
		assert(item.dwTypeData != NULL);
		if (GetMenuItemInfo(_submenu, (UINT) index, TRUE, &item) == 0)
		{
			DWORD lastError = GetLastError();
			assert(lastError == 0);
		}

		std::string itemLabel;
		Platform::Charset::toUTF8(item.dwTypeData, itemLabel);

		if (std::strcmp(itemLabel.c_str(), LABEL_DUMYITM) == 0)
		{
			// remove dummy item from submenu
			if (RemoveMenu(_submenu, index, MF_BYPOSITION) == 0)
			{
				DWORD lastError = GetLastError();
				assert(lastError == 0);
			}

			break;
		}
		else if (std::strcmp(itemLabel.c_str(), device.name().c_str()) > 0)
		{
			break;
		}
	}

	std::memset(&item, 0, sizeof(MENUITEMINFO));
	item.cbSize = sizeof(MENUITEMINFO);

	item.fMask |= MIIM_ID;
	item.wID = MAKEWORD(ID_SUBMENU_ITEM, _itemCount);

	item.fMask |= MIIM_FTYPE;
	item.fType = MFT_STRING;

	string_t name;
	Platform::Charset::fromUTF8(device.name(), name);

	item.fMask |= MIIM_STRING;
	item.dwTypeData = (LPTSTR) name.c_str();

	item.fMask |= MIIM_DATA;
	item.dwItemData = (ULONG_PTR) new DeviceInfo(device);

	item.fMask |= MIIM_STATE;
	item.fState = MFS_ENABLED;
	if (check)
		item.fState |= MFS_CHECKED;
	if (stale)
		item.fState |= MFS_GRAYED;

	// insert item into submenu
	if (InsertMenuItem(_submenu, index, TRUE, &item) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	++_itemCount;

	// scheme of using LOBYTE of item id for constant MENU_ITEM_ID and HIBYTE of
	// item id for "unique item id" works only if there are fewer than 255
	// remote speakers added; it is assumed to be a very rare case that someone
	// has more than 255 remote speakers base stations or that they are coming
	// in and out of service so frequently that the counter wraps; id 255 is
	// reserved for Options... menu item
	assert(_itemCount > 0 && _itemCount < 255);
}


bool OptionsSubmenuImpl::removeSubmenuItem(const DeviceInfo& device)
{
	bool checked = false;
	MENUITEMINFO item;

	// search submenu for item
	for (int index = 0, count = GetMenuItemCount(_submenu); index < count; ++index)
	{
		std::memset(&item, 0, sizeof(MENUITEMINFO));
		item.cbSize = sizeof(MENUITEMINFO);
		item.fMask = MIIM_STRING | MIIM_DATA | MIIM_STATE;

		if (GetMenuItemInfo(_submenu, (UINT) index, TRUE, &item) == 0)
		{
			DWORD lastError = GetLastError();
			assert(lastError == 0);
		}

		std::vector<char_t> stringBuffer(++item.cch);
		item.dwTypeData = (LPTSTR) &stringBuffer[0];
		assert(item.dwTypeData != NULL);
		if (GetMenuItemInfo(_submenu, (UINT) index, TRUE, &item) == 0)
		{
			DWORD lastError = GetLastError();
			assert(lastError == 0);
		}

		std::string itemLabel;
		Platform::Charset::toUTF8(item.dwTypeData, itemLabel);

		if (std::strcmp(itemLabel.c_str(), device.name().c_str()) == 0)
		{
			// delete item data before removing item from submenu
			delete reinterpret_cast<const DeviceInfo*>(item.dwItemData);

			// remove item from submenu
			if (RemoveMenu(_submenu, index, MF_BYPOSITION) == 0)
			{
				DWORD lastError = GetLastError();
				assert(lastError == 0);
			}

			checked = ((item.fState & MFS_CHECKED) != 0);

			break;
		}
	}

	if (GetMenuItemCount(_submenu) == 0)
	{
		// insert dummy item into submenu to indicate no devices are currently
		// advertised on the network
		std::memset(&item, 0, sizeof(MENUITEMINFO));
		item.cbSize = sizeof(MENUITEMINFO);
		item.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_STATE;
		item.fType = MFT_STRING;
		item.dwTypeData = TEXT(LABEL_DUMYITM);
		item.fState = MFS_DISABLED;
		if (InsertMenuItem(_submenu, 0, TRUE, &item) == 0)
		{
			DWORD lastError = GetLastError();
			assert(lastError == 0);
		}
	}

	return checked;
}


void OptionsSubmenuImpl::populateSubmenu(const DeviceInfoSet& devices)
{
	for (DeviceInfoSet::const_iterator it = devices.begin();
		it != devices.end(); ++it)
	{
		const DeviceInfo& device = *it;
		insertSubmenuItem(device,
			Options::getOptions()->isActivated(device.name()), device.isZeroConf());
	}
}


void OptionsSubmenuImpl::enumerateSubmenuItems(Options::SharedPtr options) const
{
	for (int index = 0, count = GetMenuItemCount(_submenu); index < count; ++index)
	{
		MENUITEMINFO item;
		std::memset(&item, 0, sizeof(MENUITEMINFO));
		item.cbSize = sizeof(MENUITEMINFO);
		item.fMask = MIIM_DATA | MIIM_STATE;

		if (GetMenuItemInfo(_submenu, (UINT) index, TRUE, &item) == 0)
		{
			DWORD lastError = GetLastError();
			assert(lastError == 0);
		}

		const bool checked = ((item.fState & MFS_CHECKED) != 0);
		const bool disabled = ((item.fState & MFS_DISABLED) != 0);
		const DeviceInfo& device =
			*(reinterpret_cast<const DeviceInfo*>(item.dwItemData));

		// save only checked or manually-created devices
		if (checked || !device.isZeroConf())
		{
			options->devices().insert(device);
			options->setActivated(device.name(), (checked && !disabled));
		}
	}
}


void OptionsSubmenuImpl::finalizeSubmenu()
{
	if (_submenu == NULL) return;

	for (int index = 0, count = GetMenuItemCount(_submenu); index < count; ++index)
	{
		MENUITEMINFO item;
		std::memset(&item, 0, sizeof(MENUITEMINFO));
		item.cbSize = sizeof(MENUITEMINFO);
		item.fMask = MIIM_DATA;

		// get menu item user data
		if (GetMenuItemInfo(_submenu, index, TRUE, &item) == 0)
		{
			DWORD lastError = GetLastError();
			assert(lastError == 0);
		}

		if (item.dwItemData != NULL)
		{
			// delete item data before removing item from submenu
			delete reinterpret_cast<const DeviceInfo*>(item.dwItemData);
		}
	}

	if (DestroyMenu(_submenu) == 0)
	{
		DWORD lastError = GetLastError();
		assert(lastError == 0);
	}

	_submenu = NULL;
}


Options::SharedPtr OptionsSubmenuImpl::buildOptions()
{
	Options::SharedPtr opts = new Options;

	// populate device info
	enumerateSubmenuItems(opts);

	const Options::SharedPtr options = Options::getOptions();

	// transfer flags
	opts->setVolumeControl(options->getVolumeControl());
	opts->setPlayerControl(options->getPlayerControl());
	opts->setResetOnPause(options->getResetOnPause());

	// transfer passwords
	for (DeviceInfoSet::const_iterator it = opts->devices().begin();
		it != opts->devices().end(); ++it)
	{
		const DeviceInfo& device = *it;
		if (!options->getPassword(device.name()).empty())
		{
			opts->setPassword(device.name(),
				options->getPassword(device.name()),
				options->getRememberPassword(device.name()));
		}
	}

	return opts;
}
