/* Copyright (c) 2013  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "Debugger.h"
#include "Platform.h"
#include "WinampPlayer.h"
#include <cassert>
#include <cstring>
#include <exception>
#include <string>
#include <vector>
#include <comdef.h>
#include <objbase.h>
#include <windows.h>

#include <mm_h.h>
#include <mm_i.c>
#include <wa_ipc.h>
#include <wasabi/api_albumart.h>
#include <wasabi/api_memmgr.h>
#include <wasabi/api_service.h>
#include <wasabi/svc_albumArtProvider.h>
#include <wasabi/svc_imgwrite.h>
#include <wasabi/waservicefactory.h>


using std::exception;


static buffer_t ToJPEG(ARGB32* const data, const int wdth, const int hght,
	api_service* const apiService, api_memmgr* const memMgmtService);


static std::string ToUTF8(const wchar_t* const wideCharString)
{
	assert(wideCharString != NULL);

	const int length = WideCharToMultiByte(CP_UTF8, 0, wideCharString, -1,
		NULL, 0, NULL, NULL);
	assert(length > 0);

	std::vector<char> multiByteString(length > 0 ? length : 1);

	const int result = WideCharToMultiByte(CP_UTF8, 0, wideCharString, -1,
		&multiByteString[0], multiByteString.size(), NULL, NULL);
	assert(length == result);

	return std::string(&multiByteString[0]);
}


static void SendMessageWithTimeout(HWND hWnd, UINT uMsg, LPARAM lPrm, WPARAM wPrm,
	DWORD* const pErr = NULL, DWORD_PTR* const pRet = NULL)
{
	if (pErr) *pErr = 0;
	if (pRet) *pRet = 0;

	if (!SendMessageTimeout(hWnd, uMsg, wPrm, lPrm, SMTO_NORMAL, 250, pRet))
	{
		if (pErr) *pErr = GetLastError();
		if (pRet) *pRet = 0;
	}
}


static void GetBasicFileInfo(time_t& length, const wchar_t* const filePath, HWND hWnd)
{
	basicFileInfoStructW fileInfo;
	fileInfo.filename = filePath;
	fileInfo.quickCheck = 0;
	fileInfo.title = NULL;
	fileInfo.titlelen = 0;
	fileInfo.length = 0;

	SendMessageWithTimeout(hWnd, WM_WA_IPC, IPC_GET_BASIC_FILE_INFOW, (WPARAM) &fileInfo);
	if (fileInfo.length > 0)
	{
		length = static_cast<time_t>(fileInfo.length) * 1000;
	}
}


static void GetTextualFileInfo(std::string& infoText, const wchar_t* const infoType,
	const wchar_t* const filePath, HWND hWnd)
{
	wchar_t wideCharString[128] = {L'\0'};

	extendedFileInfoStructW fileInfo;
	fileInfo.filename = filePath;
	fileInfo.metadata = infoType;
	fileInfo.ret = wideCharString;
	fileInfo.retlen = (sizeof(wideCharString) / sizeof(wchar_t)) - 1;

	SendMessageWithTimeout(hWnd, WM_WA_IPC, IPC_GET_EXTENDED_FILE_INFOW, (WPARAM) &fileInfo);
	if (fileInfo.ret[0] != L'\0')
	{
		infoText = ToUTF8(fileInfo.ret);
	}
}


static void GetVisualFileInfo(buffer_t& imageData, std::string& imageType,
	const wchar_t* const infoType, const wchar_t* const filePath,
	api_service* const apiService, api_memmgr* const memMgmtService)
{
	const size_t n = apiService->service_getNumServices(
						svc_albumArtProvider::getServiceType());
	for (size_t i = 0; i < n; ++i)
	{
		waServiceFactory* const aapServiceFactory =
			apiService->service_enumService(svc_albumArtProvider::getServiceType(), i);
		if (aapServiceFactory != NULL)
		{
			svc_albumArtProvider* const aapService =
				(svc_albumArtProvider*) aapServiceFactory->getInterface();
			if (aapService != NULL)
			{
				void*     data = 0;
				size_t    size = 0;
				wchar_t*  type = 0;
				const int code = aapService->GetAlbumArtData(
					filePath, infoType, &data, &size, &type);
				if (code == ALBUMARTPROVIDER_SUCCESS && data != 0)
				{
					if (size != 0 && type != 0)
					{
						imageData.resize(size);
						std::memcpy(&imageData[0], data, size);
						imageType = "image/" + ToUTF8(type);

						i = n; // break out of loop after cleanup
					}

					if (data) memMgmtService->sysFree(data);
					if (type) memMgmtService->sysFree(type);
				}

				aapServiceFactory->releaseInterface(aapService);
			}

			apiService->service_release(aapServiceFactory);
		}
	}
}


static void GetVisualFileInfo2(buffer_t& imageData, std::string& imageType,
	const wchar_t* const infoType, const wchar_t* const filePath,
	api_service* const apiService, api_memmgr* const memMgmtService)
{
	waServiceFactory* const aaServiceFactory =
		apiService->service_getServiceByGuid(albumArtGUID);
	if (aaServiceFactory != NULL)
	{
		api_albumart* const aaService =
			(api_albumart*) aaServiceFactory->getInterface();
		if (aaService != NULL)
		{
			int       wdth = 0;
			int       hght = 0;
			ARGB32*   data = 0;
			const int code = aaService->GetAlbumArt(
				filePath, infoType, &wdth, &hght, &data);
			if (code == ALBUMART_SUCCESS && data != 0)
			{
				if (wdth > 0 && hght > 0)
				{
					imageData = ToJPEG(data, wdth, hght, apiService, memMgmtService);
					imageType = "image/jpeg";
				}

				if (data) memMgmtService->sysFree(data);
			}

			aaServiceFactory->releaseInterface(aaService);
		}

		apiService->service_release(aaServiceFactory);
	}
}


static buffer_t ToJPEG(ARGB32* const data, const int wdth, const int hght,
	api_service* const apiService, api_memmgr* const memMgmtService)
{
	buffer_t jpeg(0);

	const size_t n = apiService->service_getNumServices(
						svc_imageWriter::getServiceType());
	for (size_t i = 0; i < n; ++i)
	{
		waServiceFactory* const imgServiceFactory =
			apiService->service_enumService(svc_imageWriter::getServiceType(), i);
		if (imgServiceFactory != NULL)
		{
			svc_imageWriter* const imgService =
				(svc_imageWriter*) imgServiceFactory->getInterface();
			if (imgService != NULL)
			{
				if (imgService->bitDepthSupported(32)
					&& imgService->getImageTypeName() != NULL
					&& imgService->getImageTypeName() == std::wstring(L"JPEG"))
				{
					int   jpegSize = 0;
					void* jpegData = imgService->convert(data, 32, wdth, hght, &jpegSize);
					if (jpegData != NULL && jpegSize > 0)
					{
						jpeg.resize(jpegSize);
						std::memcpy(&jpeg[0], jpegData, jpegSize);
					}

					if (jpegData) memMgmtService->sysFree(jpegData);
				}

				imgServiceFactory->releaseInterface(imgService);
			}

			apiService->service_release(imgServiceFactory);
		}
	}

	return jpeg;
}


struct hresult_error : exception
{
	hresult_error(HRESULT result)
	: exception(hresult_string(result).c_str())
	{
	}
	hresult_error(HRESULT result, const char* const context)
	: exception(hresult_string(result).append(" from ").append(context).c_str())
	{
	}
	static std::string hresult_string(HRESULT result)
	{
		return ToUTF8(_com_error(result).ErrorMessage());
	}
};


//------------------------------------------------------------------------------


WinampPlayer::WinampPlayer(HWND& playerWindow)
:
	_apiService(NULL),
	_memMgmtService(NULL),
	_memMgmtServiceFactory(NULL),
	_playerWindow(playerWindow)
{
}


void WinampPlayer::play()
{
	// send 'play' command to player
	sendCommand(40045);
}


void WinampPlayer::pause()
{
	// send 'pause' command to player
	sendCommand(40046);
}


void WinampPlayer::stop()
{
	// send 'stop' command to player
	sendCommand(40047);
}


void WinampPlayer::restart()
{
	const uint8_t status = reinterpret_cast<uint8_t>(sendMessage(IPC_ISPLAYING));

	// check if player is playing
	if (status > 0)
	{
		// send 'reset playback position' message to player
		sendMessage(IPC_JUMPTOTIME, 0);
	}

	// check if player is paused
	if (status == 3)
	{
		// resume playback
		play();
	}
}


void WinampPlayer::startNext()
{
	// send 'skip to next track' command to player
	sendCommand(40048);
}


void WinampPlayer::startPrev()
{
	// send 'skip to previous track' command to player
	sendCommand(40044);
}


void WinampPlayer::increaseVolume()
{
	// send 'volume up' command to player (twice)
	sendCommand(40058);
	sendCommand(40058);
}


void WinampPlayer::decreaseVolume()
{
	// send 'volume down' command to player (twice)
	sendCommand(40059);
	sendCommand(40059);
}


void WinampPlayer::toggleMute()
{
	static uint8_t previousVolume = 0;

	// send 'get volume' message to player
	const uint8_t volume = reinterpret_cast<uint8_t>(sendMessage(IPC_SETVOLUME, -666));

	if (volume > 0)
	{
		// send 'mute volume' message to player
		sendMessage(IPC_SETVOLUME, 0);
	}
	else
	{
		// send 'set volume' message to player
		sendMessage(IPC_SETVOLUME, previousVolume);
	}

	previousVolume = volume;
}


void WinampPlayer::toggleShuffle()
{
	// send 'get shuffle state' message to player
	const int state = reinterpret_cast<int>(sendMessage(IPC_GET_SHUFFLE));
	// send 'set shuffle state' message to player
	sendMessage(IPC_SET_SHUFFLE, !state);
}


//------------------------------------------------------------------------------


#define DO(exp) {HRESULT hr(exp); if (FAILED(hr)) throw hresult_error(hr,#exp);}


time_t WinampPlayer::getPlaybackMetadata(std::string& title, std::string& album,
	std::string& artist, buffer_t& artworkData, std::string& artworkType) const
{
	time_t length = 0;
	title.clear(), album.clear(), artist.clear();
	artworkData.resize(0), artworkType.assign("image/none");

	const wchar_t* const filePath =
		static_cast<wchar_t*>(sendMessage(IPC_GET_PLAYING_FILENAME));
	if (filePath != NULL && filePath[0] != L'\0')
	{
		GetBasicFileInfo(length, filePath, window());
		GetTextualFileInfo(title, L"Title", filePath, window());
		GetTextualFileInfo(album, L"Album", filePath, window());
		GetTextualFileInfo(artist, L"Artist", filePath, window());

		if (_apiService == NULL)
		{
			// try to initialize player service interfaces
			const_cast<WinampPlayer*>(this)->initServices();
		}
		if (_apiService != NULL && _memMgmtService != NULL)
		{
			GetVisualFileInfo(artworkData, artworkType, L"cover",
							filePath, _apiService, _memMgmtService);

			if (artworkData.empty())
			{
				// try an alternate API for getting artwork
				GetVisualFileInfo2(artworkData, artworkType, L"cover",
								filePath, _apiService, _memMgmtService);
			}
		}
	}

	if (length == 0)
	{
		// try an alternate API for getting track length
		length = reinterpret_cast<time_t>(sendMessage(IPC_GETOUTPUTTIME, 2));
	}

	if (title.empty())
	{
		// try an alternate API for getting track title
		const wchar_t* const wideCharTitle =
			static_cast<wchar_t*>(sendMessage(IPC_GET_PLAYING_TITLE));
		if (wideCharTitle != NULL && wideCharTitle[0] != L'\0')
		{
			title = ToUTF8(wideCharTitle);
		}
	}

	if ((length == 0 || title.empty()) && FindWindow(L"TFMainWindow", NULL))
	{
		// try MediaMonkey's COM API for getting track metadata

		ISDBApplication * sdbApplication  = NULL;
		ISDBPlayer      * sdbPlayer       = NULL;
		ISDBSongData    * sdbSongData     = NULL;
		ISDBAlbumArtList* sdbAlbumArtList = NULL;
		ISDBAlbumArtItem* sdbAlbumArtItem = NULL;
		ISDBImage       * sdbImage        = NULL;

		BSTR imageType  = NULL;
		BSTR songTitle  = NULL;
		BSTR songAlbum  = NULL;
		BSTR songArtist = NULL;

		CoInitialize(NULL);

		try
		{
			DO (CoCreateInstance(CLSID_SDBApplication, NULL, CLSCTX_LOCAL_SERVER,
								 IID_ISDBApplication, (LPVOID*) &sdbApplication));

			DO (sdbApplication->get_Player(&sdbPlayer));

			DO (sdbPlayer->get_CurrentSong(&sdbSongData));

			long songLength = 0;
			DO (sdbSongData->get_SongLength(&songLength));
			if (songLength > 0) length = songLength;

			DO (sdbSongData->get_Title(&songTitle));
			if (songTitle != NULL && songTitle[0] != L'\0') title = ToUTF8(songTitle);

			DO (sdbSongData->get_AlbumName(&songAlbum));
			if (songAlbum != NULL && songAlbum[0] != L'\0') album = ToUTF8(songAlbum);

			DO (sdbSongData->get_ArtistName(&songArtist));
			if (songArtist != NULL && songArtist[0] != L'\0') artist = ToUTF8(songArtist);

			DO (sdbSongData->get_AlbumArt(&sdbAlbumArtList));

			long imageCount = 0;
			DO (sdbAlbumArtList->get_Count(&imageCount));
			if (imageCount > 0)
			{
				DO (sdbAlbumArtList->get_Item(0, &sdbAlbumArtItem));

				DO (sdbAlbumArtItem->get_Image(&sdbImage));

				long imageData = 0;
				DO (sdbImage->get_ImageData(&imageData));

				long imageSize = 0;
				DO (sdbImage->get_ImageDataLen(&imageSize));

				DO (sdbImage->get_ImageFormat(&imageType));
				if (imageType != NULL && imageType[0] != L'\0')
				{
					artworkData.resize(imageSize);
					std::memcpy(&artworkData[0], (void*) imageData, imageSize);
					artworkType = ToUTF8(imageType);
				}
			}
		}
		CATCH_ALL

		if (imageType != NULL)  SysFreeString(imageType);
		if (songTitle != NULL)  SysFreeString(songTitle);
		if (songAlbum != NULL)  SysFreeString(songAlbum);
		if (songArtist != NULL) SysFreeString(songArtist);

		if (sdbImage != NULL)        sdbImage->Release();
		if (sdbAlbumArtItem != NULL) sdbAlbumArtItem->Release();
		if (sdbAlbumArtList != NULL) sdbAlbumArtList->Release();
		if (sdbSongData != NULL)     sdbSongData->Release();
		if (sdbPlayer != NULL)       sdbPlayer->Release();
		if (sdbApplication != NULL)  sdbApplication->Release();

		CoUninitialize();
	}

	return length;
}


//------------------------------------------------------------------------------


void WinampPlayer::freeServices()
{
	if (_apiService != NULL)
	{
		if (_memMgmtServiceFactory != NULL)
		{
			if (_memMgmtService != NULL)
			{
				_memMgmtServiceFactory->releaseInterface(_memMgmtService);
				_memMgmtService = NULL;
			}

			_apiService->service_release(_memMgmtServiceFactory);
			_memMgmtServiceFactory = NULL;
		}

		_apiService = NULL;
	}
}


void WinampPlayer::initServices()
{
	_apiService = static_cast<api_service*>(sendMessage(IPC_GET_API_SERVICE));
	if (_apiService == reinterpret_cast<api_service*>(1))
	{
		// service interface not supported
		_apiService = NULL;
	}
	else if (_apiService != NULL)
	{
		_memMgmtServiceFactory = _apiService->service_getServiceByGuid(memMgrApiServiceGuid);
		if (_memMgmtServiceFactory == NULL)
		{
			freeServices();
			return;
		}

		_memMgmtService = static_cast<api_memmgr*>(_memMgmtServiceFactory->getInterface());
		if (_memMgmtService == NULL)
		{
			freeServices();
			return;
		}
	}
}


void WinampPlayer::sendCommand(const int param) const
{
	DWORD error;
	DWORD_PTR result;

	assert(param > 0);
	assert(window() > 0);

	SendMessageWithTimeout(window(), WM_COMMAND, (LPARAM) 0, (WPARAM) param, &error, &result);

	if (error == ERROR_TIMEOUT)
	{
		Debugger::printf("WinampPlayer::sendCommand(%i) timed out", param);
	}
	else if (error)
	{
		Debugger::printf("WinampPlayer::sendCommand(%i) failed with error code %lu", param, error);
	}
	else if (result)
	{
		Debugger::printf("WinampPlayer::sendCommand(%i) not accepted by target", param);
	}
}


void* WinampPlayer::sendMessage(const long lParam, const int wParam) const
{
	DWORD error;
	DWORD_PTR result;

	assert(lParam > 0);
	assert(window() > 0);

	SendMessageWithTimeout(window(), WM_WA_IPC, (LPARAM) lParam, (WPARAM) wParam, &error, &result);

	if (error == ERROR_TIMEOUT)
	{
		Debugger::printf("WinampPlayer::sendMessage(%li, %i) timed out", lParam, wParam);
	}
	else if (error)
	{
		Debugger::printf("WinampPlayer::sendMessage(%li, %i) failed with error code %lu", lParam, wParam, error);
	}

	return reinterpret_cast<void*>(result);
}
