/* Copyright (c) 2013  Eric Milles <eric.milles@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "XMPlayPlayer.h"
#include <cstdio>


XMPlayPlayer::XMPlayPlayer(XMPFUNC_MISC* miscFunc, XMPFUNC_STATUS* statusFunc)
:
	_miscFunc(miscFunc),
	_statusFunc(statusFunc)
{
}


BYTE XMPlayPlayer::majorVersion() const
{
	DWORD playerVersion = _miscFunc->GetVersion();
	BYTE majorVersion = HIBYTE(HIWORD(playerVersion));
	return majorVersion;
}


BYTE XMPlayPlayer::minorVersion() const
{
	DWORD playerVersion = _miscFunc->GetVersion();
	BYTE minorVersion = LOBYTE(HIWORD(playerVersion));
	return minorVersion;
}


HWND XMPlayPlayer::window() const
{
	return _miscFunc->GetWindow();
}


void XMPlayPlayer::play()
{
	_miscFunc->PerformShortcut(80); // play/pause
}


void XMPlayPlayer::pause()
{
	_miscFunc->PerformShortcut(80); // play/pause
}


DWORD WINAPI stopImpl(LPVOID context)
{
	return ((XMPFUNC_MISC*) context)->PerformShortcut(81); // stop
}

void XMPlayPlayer::stop()
{
	// stop shortcut is executed in calling thread and triggers Output::close;
	// perform shortcut asynchronously in worker thread to avoid deadlock of
	// thread waiting for itself to exit
	QueueUserWorkItem(stopImpl, _miscFunc, WT_EXECUTEDEFAULT);
}


DWORD WINAPI restartImpl(LPVOID context)
{
	return ((XMPFUNC_MISC*) context)->PerformShortcut(84); // restart track
}

void XMPlayPlayer::restart()
{
	// restart shortcut is executed in calling thread and triggers Output::close;
	// perform shortcut asynchronously in worker thread to avoid deadlock of
	// thread waiting for itself to exit
	QueueUserWorkItem(restartImpl, _miscFunc, WT_EXECUTEDEFAULT);
}


void XMPlayPlayer::startNext()
{
	_miscFunc->PerformShortcut(128); // next track
}


void XMPlayPlayer::startPrev()
{
	_miscFunc->PerformShortcut(129); // previous track
}


void XMPlayPlayer::increaseVolume()
{
	_miscFunc->PerformShortcut(512); // volume up
	_miscFunc->PerformShortcut(512); // each call produces a 2% change in volume
}


void XMPlayPlayer::decreaseVolume()
{
	_miscFunc->PerformShortcut(513); // volume down
	_miscFunc->PerformShortcut(513); // each call produces a 2% change in volume
}


void XMPlayPlayer::toggleMute()
{
	_miscFunc->PerformShortcut(523); // toggle volume mute
}


void XMPlayPlayer::toggleShuffle()
{
	_miscFunc->PerformShortcut(313); // toggle random play order
}


time_t XMPlayPlayer::getPlaybackMetadata(
	std::string& title, std::string& album, std::string& artist) const
{
	time_t length = 0;
	title.clear(), album.clear(), artist.clear();

	const std::string infoString(_miscFunc->GetInfoText(XMPINFO_TEXT_GENERAL));

	std::string::size_type beg = 0;
	while (beg != std::string::npos)
	{
		std::string::size_type mid = infoString.find_first_of('\t', beg);
		std::string::size_type end = infoString.find_first_of('\n', beg);

		if (mid != std::string::npos && mid < end)
		{
			const std::string key(infoString.substr(beg, mid - beg));
			++mid; // advance past '\t'
			const std::string val(infoString.substr(mid, end - mid));

			if (key == "Length")
			{
				unsigned int mins = 0, secs = 0;
				sscanf_s(val.c_str(), "%u:%u", &mins, &secs);
				// calculate length in milliseconds
				length = (mins * 60000) + (secs * 1000);
			}
			else if (key == "Title" || (key == "File" && title.empty()))
			{
				title = val;
			}
		}

		beg = infoString.find_first_not_of('\n', end);
	}

	const char* tag;
	if ((tag = _miscFunc->GetTag(2)) != NULL)
	{
		album = tag;
	}
	if ((tag = _miscFunc->GetTag(1)) != NULL)
	{
		artist = tag;
	}

	return length;
}


time_t XMPlayPlayer::getPlaybackPosition() const
{
	const double offset = _statusFunc->GetTime();
	// convert offset from seconds to milliseconds
	return static_cast<time_t>(offset * 1000.0);
}
